import string
import re
from re import Pattern
from datetime import datetime
from abc import ABC, ABCMeta, abstractmethod
from typing import Callable

from api.database import *
from api.miscellaneous import *


class ValueType(ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self, api_id: string, may_be_none: bool):
        self.api_id = api_id
        self._may_be_none = may_be_none
    
    def add_restrictions_to(self, configuration_entry: dict):
        if self._may_be_none:
            configuration_entry["may_be_null"] = self._may_be_none
    
    @abstractmethod
    def db_value_to_json(self, db_value: DbValueType) -> JsonTypes:
        pass  # pragma: no cover
    
    @abstractmethod
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> DbValueType:
        pass  # pragma: no cover


class SimpleType(ValueType):
    
    def __init__(self, api_id: string,
                 db_to_json: Callable[[DbValueType], JsonTypes],
                 client_to_db: Callable[[CJsonValue], DbValueType],
                 may_be_none: bool = False):
        super().__init__(api_id, may_be_none)
        self.db_to_json = db_to_json
        self.client_to_db = client_to_db
    
    def db_value_to_json(self, db_value: DbValueType) -> JsonTypes:
        return self.db_to_json(db_value)
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> DbValueType:
        return self.client_to_db(client_value)


class ObjectIdType(ValueType):
    
    def __init__(self,
                 api_id: string,
                 may_be_none: bool,
                 db_table: str,
                 id_db_column: str,
                 check_deleted: bool,
                 non_existent_message: str):
        super().__init__(api_id, may_be_none)
        self._db_table = db_table
        self._id_db_column = id_db_column
        self._non_existent_message = non_existent_message
        
        db_ensure_valid_id(db_table)
        db_ensure_valid_id(id_db_column)
        
        self._query_object_statement = PreparedStatement(f"""
            SELECT "{id_db_column}"
            FROM "{db_table}"
            WHERE "{id_db_column}" = ?
                {'AND (NOT "deleted")' if check_deleted else ""}
            """)
    
    def db_value_to_json(self, db_value: int or str or None) -> int or None:
        if self._may_be_none and db_value is None:
            return None
        return int(db_value)
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> int or None:
        if self._may_be_none and client_value.is_null():
            return None
        id = client_value.as_int(0, MAX_VALUE_SINT32)
        if len(transaction.execute_statement(self._query_object_statement, id)) == 0:
            raise ApiClientException(ERROR_OBJECT_ERROR(self._non_existent_message))
        return id


class NumberType(ValueType):
    __metaclass__ = ABCMeta
    
    def __init__(self, api_id: string,
                 absolute_min: int, absolute_max: int,
                 min_value: int = None, max_value: int = None):
        super().__init__(api_id, False)
        self.has_custom_min = min_value is not None
        self.has_custom_max = max_value is not None
        if min_value is None:
            min_value = absolute_min
        if max_value is None:
            max_value = absolute_max
        assert absolute_min <= min_value <= max_value <= absolute_max
        self.min_value = min_value
        self.max_value = max_value
    
    def add_restrictions_to(self, configuration_entry: dict):
        super().add_restrictions_to(configuration_entry)
        if self.has_custom_min:
            configuration_entry["min_value"] = self.min_value
        if self.has_custom_max:
            configuration_entry["max_value"] = self.max_value
    
    def db_value_to_json(self, db_value: int or None) -> int or None:
        return db_value
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> int or None:
        if self._may_be_none and client_value.is_null():
            return None
        return client_value.as_int(self.min_value, self.max_value)


class IntType(NumberType):
    def __init__(self, min_value: int = None, max_value: int = None):
        super().__init__("int", MIN_VALUE_SINT32, MAX_VALUE_SINT32, min_value, max_value)


class LongType(NumberType):
    def __init__(self, min_value: int = None, max_value: int = None):
        super().__init__("long", MIN_VALUE_SINT64, MAX_VALUE_SINT64, min_value, max_value)


class NumberToStringType(NumberType):
    
    def __init__(self,
                 api_id: string,
                 absolute_min: int,
                 absolute_max: int,
                 min_value: int = None,
                 max_value: int = None):
        super().__init__(api_id, absolute_min, absolute_max, min_value, max_value)
    
    def db_value_to_json(self, db_value: str or None) -> JsonTypes:
        if db_value is None:
            return None
        return int(db_value)
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> DbValueType:
        if client_value.is_null() and self._may_be_none:
            return None
        int_val = client_value.as_int(self.min_value, self.max_value)
        return str(int_val)


class IntToStringType(NumberToStringType):
    def __init__(self, min_value: int = None, max_value: int = None):
        super().__init__("int", MIN_VALUE_SINT32, MAX_VALUE_SINT32, min_value, max_value)


class LongToStringType(NumberToStringType):
    def __init__(self, min_value: int = None, max_value: int = None):
        super().__init__("long", MIN_VALUE_SINT64, MAX_VALUE_SINT64, min_value, max_value)


class AbstractStringType(ValueType, ABC):
    """
    Base class for all types which use a string as the json value
    """
    __metaclass__ = ABCMeta
    
    def __init__(self, api_id: string, may_be_none: bool):
        super().__init__(api_id, may_be_none)
    
    @abstractmethod
    def get_max_string_length(self) -> int:
        """
        Maximum length any valid json value might have (It may be greater than the actual maximum)
        """
        pass  # pragma: no cover


class SimpleStringType(AbstractStringType):
    
    def __init__(self,
                 api_id: string,
                 max_length: int,
                 min_length: int or None,
                 regex: str or None,
                 may_be_none: bool = False):
        super().__init__(api_id, may_be_none)
        self._min_length = min_length
        self._max_length = max_length
        self._pattern: Pattern or None = None
        if regex is not None:
            self._pattern = re.compile(regex)
    
    def get_max_string_length(self) -> int:
        """Returns the exact maximum length of this string"""
        return self._max_length
    
    def add_restrictions_to(self, configuration_entry: dict):
        super().add_restrictions_to(configuration_entry)
        if self.api_id == "string" or self.api_id == "url":
            if self._min_length is not None:
                configuration_entry["min_length"] = self._min_length
            configuration_entry["max_length"] = self._max_length
    
    def db_value_to_json(self, db_value: str or None) -> str or None:
        return db_value
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> str or None:
        if self._may_be_none and client_value.is_null():
            return None
        value = client_value.as_string(self._max_length, self._min_length)
        if self._pattern is None:
            return value
        if self._pattern.fullmatch(value) is None:
            client_value.raise_error("String does not match pattern")
        return value


class UniqueStringType(SimpleStringType):
    
    def __init__(self,
                 api_id: string,
                 max_length: int,
                 min_length: int or None,
                 regex: str or None,
                 db_table: str,
                 id_db_column: str,
                 value_db_column: str,
                 may_be_none: bool = False):
        super().__init__(api_id, max_length, min_length, regex, may_be_none)
        
        db_ensure_valid_id(db_table)
        db_ensure_valid_id(id_db_column)
        db_ensure_valid_id(value_db_column)
        
        self._request_statement = PreparedStatement(f"""
            SELECT "{value_db_column}" FROM "{db_table}"
            WHERE "{value_db_column}" = ?
                AND (NOT "{id_db_column}" = ?)
            """)
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> str or None:
        value = super().validate_and_convert_client_value(transaction, object_id, client_value)
        if len(transaction.execute_statement(self._request_statement, value, object_id)) > 0:
            raise ApiClientException(ERROR_OBJECT_ERROR(f"The value {truncate_string(value)} is already used by "
                                                        f"another object (Which may have been deleted)"))
        return value


class StringType(SimpleStringType):
    
    def __init__(self, max_length: int, min_length: int = None, may_be_none: bool = False):
        super().__init__("string", max_length, min_length, None, may_be_none)


class MappedStringType(AbstractStringType):
    
    def __init__(self, api_id: string, may_be_none: bool, json_to_db: dict[str, str or int]):
        super().__init__(api_id, may_be_none)
        self.__json_to_db = json_to_db
        self.__db_to_json = {}
        self.__max_json_value_length = 100
        
        for json, db in json_to_db.items():
            self.__max_json_value_length = max(self.__max_json_value_length, len(json))
            if db in self.__db_to_json:
                raise ValueError(f"Duplicated database value '{db}'")  # pragma: no cover
            self.__db_to_json[db] = json
    
    def get_max_string_length(self) -> int:
        return self.__max_json_value_length
    
    def db_value_to_json(self, db_value: str or int or None) -> str or None:
        if self._may_be_none and db_value is None:
            return None
        return self.__db_to_json[db_value]
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> str or int or None:
        if self._may_be_none and client_value.is_null():
            return None
        string = client_value.as_string(self.__max_json_value_length)
        if string not in self.__json_to_db:
            client_value.raise_error("Unknown value")
        return self.__json_to_db[string]


SEMESTER_STRING_REGEX = "([0-9]{4}(ws|ss)|none)"
SEMESTER_STRING_PATTERN = re.compile(SEMESTER_STRING_REGEX)


class SemesterStringType(AbstractStringType):
    
    def __init__(self):
        super().__init__("semester_string", False)
    
    def get_max_string_length(self) -> int:
        """Returns the exact maximum length of this string"""
        return 6
    
    def db_value_to_json(self, db_value: str) -> str:
        if len(db_value) == 0 or db_value.isspace():
            return "none"
        if SEMESTER_STRING_PATTERN.fullmatch(db_value) is None:
            print(f"Warning: Invalid semester string in database: {truncate_string(db_value)}")
            return "none"
        return db_value
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> str or None:
        value = client_value.as_string(50, 0)
        if SEMESTER_STRING_PATTERN.fullmatch(value) is None:
            client_value.raise_error("String does not match pattern")
        if value == "none":
            value = ""
        return value
    

class DatetimeType(AbstractStringType):
    
    def __init__(self, may_be_none: bool = False):
        super().__init__("datetime", may_be_none)
    
    def get_max_string_length(self) -> int:
        return 100
    
    def db_value_to_json(self, db_value: datetime or None) -> str or None:
        if self._may_be_none and db_value is None:
            return None
        return db_value.strftime(API_DATETIME_FORMAT)
    
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> datetime or None:
        if self._may_be_none and client_value.is_null():
            return None
        time_string = client_value.as_string(100)
        try:
            return datetime.strptime(time_string, API_DATETIME_FORMAT)
        except ValueError:
            client_value.raise_error("Invalid datetime format")
