from flask import request

from api.routes import *
from api.course import (COURSE_SECONDARY_DB_SELECTION, COURSE_SECONDARY_COLUMN_DOWNLOADABLE,
                        COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE, COURSE_SECONDARY_COLUMN_VISIBLE,
                        course_list_db_to_json_no_lectures, course_secondary_db_to_json_no_lectures,
                        course_db_to_json_no_lectures, course_queue_query_auth,
                        lecture_db_to_json_no_chapters_media, lecture_queue_query_auth,
                        chapter_db_to_json, media_source_with_format_db_to_json,
                        course_queue_search, lecture_queue_search)
import api

_SQL_GET_ALL_COURSES = PreparedStatement("""
SELECT * FROM "courses"
WHERE ? OR (
            "visible"
        AND "listed"
    )
ORDER BY "id" ASC
""")
_SQL_GET_ALL_COURSES_AUTH = PreparedStatement("""
SELECT "perm".* FROM "perm"
JOIN "courses" ON ("courses"."id" = "perm"."course_id")
WHERE ( ? OR (
        "courses"."visible"
        AND "courses"."listed"
    ))
    AND (NOT "perm"."deleted")
ORDER BY "courses"."id" ASC
""")


@api_route("/courses", ["GET"], allow_while_readonly=True)
def api_route_courses():
    is_mod: bool = is_moderator()
    courses_db, auth_db = db_pool.execute_read_statements_in_transaction(
        (_SQL_GET_ALL_COURSES, [is_mod]),
        (_SQL_GET_ALL_COURSES_AUTH, [is_mod])
    )
    courses_list_json = course_list_db_to_json_no_lectures(courses_db, auth_db, is_mod)
    courses_map_json = {}
    for course_json in courses_list_json:
        courses_map_json[str(course_json["id"])] = course_json
    
    return {
        "courses": courses_map_json
    }


_SQL_GET_COURSE_BY_ID_NO_CHECK = PreparedStatement("""
SELECT * FROM "courses"
WHERE "id"= ?
""")
_SQL_GET_COURSE_BY_ID_STRING_NO_CHECK = PreparedStatement("""
SELECT * FROM "courses"
WHERE "handle"= ?
""")
_SQL_GET_COURSE_LECTURES = PreparedStatement("""
SELECT * FROM "lectures"
WHERE "course_id" = ?
    AND ( ? OR "visible")
ORDER BY "id" ASC
""")
_SQL_GET_COURSE_LECTURES_AUTH = PreparedStatement("""
SELECT "perm".* FROM "perm"
LEFT JOIN "lectures" ON ("perm"."lecture_id" = "lectures"."id")
WHERE
    "lectures"."course_id" = ?
    AND ( ? OR "lectures"."visible")
    AND (NOT "perm"."deleted")
ORDER BY "lectures"."id" ASC
""")
_SQL_GET_COURSE_LECTURES_CHAPTERS = PreparedStatement("""
SELECT "chapters".* FROM "chapters"
JOIN "lectures" ON ("lectures"."id" = "chapters"."lecture_id")
WHERE "lectures"."course_id" = ?
    AND (NOT "chapters"."deleted")
    AND ( ? OR (
            "chapters"."visible"
        AND "lectures"."visible"
    ))
ORDER BY "lectures"."id" ASC, "chapters"."time" ASC
""")
_SQL_GET_COURSE_LECTURES_MEDIA = PreparedStatement("""
SELECT "videos".*,
    "formats"."description" AS "format_description",
    "formats"."resolution" AS "format_resolution",
    "formats"."aspect" AS "format_aspect",
    "formats"."player_prio" AS "format_player_prio"
FROM "videos"
JOIN "formats" ON ("videos"."video_format" = "formats"."id")
JOIN "lectures" ON ("lectures"."id" = "videos"."lecture_id")
WHERE "lectures"."course_id" = ?
    AND ( ? OR (
            "videos"."visible"
        AND "lectures"."visible"
    ))
ORDER BY "lectures"."id" ASC
""")


def _db_execute_get_courses(trans: ReadTransaction,
                            is_mod: bool,
                            course_id: int = None,
                            course_id_string: str = None) -> tuple[
    DbResult, DbResultSet, DbResultSet or None, DbResultSet or None, DbResultSet or None, DbResultSet or None
]:
    if course_id:
        courses_array = trans.execute_statement(_SQL_GET_COURSE_BY_ID_NO_CHECK, course_id)
    else:
        courses_array = trans.execute_statement(_SQL_GET_COURSE_BY_ID_STRING_NO_CHECK, course_id_string)
    
    if len(courses_array) == 0:
        trans.close()
        raise ApiClientException(ERROR_UNKNOWN_OBJECT)
    if len(courses_array) > 1:  # pragma: no cover
        raise Exception("Got more than one course with id " + str(course_id if course_id else course_id_string))
    course_db = courses_array[0]
    if not is_mod and not bool(course_db["visible"]):
        raise ApiClientException(ERROR_UNAUTHORIZED)
    
    course_id = course_db["id"]
    
    auth_stat = course_queue_query_auth(trans, course_id)
    if "include_lectures" not in request.args or request.args["include_lectures"] != "true":
        trans.close()
        return course_db, trans.get_result(auth_stat), None, None, None, None
    
    lectures_db, lectures_auth_db, chapters_db, media_db = trans.execute_statements_and_close(
        (_SQL_GET_COURSE_LECTURES, [course_id, is_mod]),
        (_SQL_GET_COURSE_LECTURES_AUTH, [course_id, is_mod]),
        (_SQL_GET_COURSE_LECTURES_CHAPTERS, [course_id, is_mod]),
        (_SQL_GET_COURSE_LECTURES_MEDIA, [course_id, is_mod]),
    )
    return course_db, trans.get_result(auth_stat), lectures_db, lectures_auth_db, chapters_db, media_db


@api_add_route("/course/<int:course_id>", ["GET"])
@api_add_route("/course/<string:course_id_string>", ["GET"])
@api_function(allow_while_readonly=True)
def api_route_course(course_id: int = None, course_id_string: str = None):
    is_mod: bool = is_moderator()
    
    course_db, course_auth_db, lectures_db, lectures_auth_db, chapters_db, media_db = db_pool.execute_read_transaction(
        _db_execute_get_courses,
        is_mod,
        course_id,
        course_id_string
    )
    course_json = course_db_to_json_no_lectures(course_db, course_auth_db, is_mod)
    if lectures_db is None:
        return course_json
    
    lectures_auth_i = 0
    chapters_i = 0
    media_i = 0
    
    lectures_json = []
    course_json["lectures"] = lectures_json
    for lecture_db in lectures_db:
        lecture_id: int = lecture_db["id"]
        lectures_auth_i, lecture_auth_db = db_collect_id_sorted_data(
            lectures_auth_db, lectures_auth_i,
            "lecture_id", lecture_id)
        lecture_json = lecture_db_to_json_no_chapters_media(lecture_db,
                                                            course_auth_db + lecture_auth_db,
                                                            not bool(course_db["embedinvisible"]),
                                                            is_mod)
        lectures_json.append(lecture_json)
        
        chapters_i, lecture_json["chapters"] = db_collect_id_sorted_data(
            chapters_db, chapters_i,
            "lecture_id", lecture_id,
            lambda c: chapter_db_to_json(c, is_mod))
        
        media_i, lecture_json["media_sources"] = db_collect_id_sorted_data(
            media_db, media_i,
            "lecture_id", lecture_id,
            lambda m: media_source_with_format_db_to_json(m, bool(course_db["downloadable"]), is_mod))
    return course_json


_SQL_GET_LECTURE_WITH_COURSE_NO_CHECK = PreparedStatement(f"""
SELECT "lectures".*, {COURSE_SECONDARY_DB_SELECTION}
FROM "lectures"
JOIN "courses" ON ("courses"."id" = "lectures"."course_id")
WHERE "lectures"."id" = ?
""")
_SQL_GET_LECTURE_CHAPTERS_NO_ID_COND = """
SELECT * FROM "chapters"
WHERE %s
    AND (NOT "chapters"."deleted")
    AND ( ? OR "chapters"."visible")
ORDER BY "time" ASC
"""
_SQL_GET_LECTURE_CHAPTERS = PreparedStatement(_SQL_GET_LECTURE_CHAPTERS_NO_ID_COND % '"chapters"."lecture_id" = ?')
_SQL_GET_LECTURE_MEDIA_NO_ID_COND = """
SELECT "videos".*,
    "formats"."description" AS "format_description",
    "formats"."resolution" AS "format_resolution",
    "formats"."aspect" AS "format_aspect",
    "formats"."player_prio" AS "format_player_prio"
FROM "videos"
JOIN "formats" ON ("videos"."video_format" = "formats"."id")
WHERE %s
    AND ( ? OR "videos"."visible")
"""
_SQL_GET_LECTURE_MEDIA = PreparedStatement(_SQL_GET_LECTURE_MEDIA_NO_ID_COND % '"videos"."lecture_id" = ?')


def _db_execute_get_lecture(trans: ReadTransaction, is_mod: bool, lecture_id: int) -> tuple[
    DbResultSet, DbResultSet, DbResultSet, DbResultSet
]:
    auth_stat = lecture_queue_query_auth(trans, lecture_id)
    lecture_list_db, chapters_db, media_db = trans.execute_statements_and_close(
        (_SQL_GET_LECTURE_WITH_COURSE_NO_CHECK, [lecture_id]),
        (_SQL_GET_LECTURE_CHAPTERS, [lecture_id, is_mod]),
        (_SQL_GET_LECTURE_MEDIA, [lecture_id, is_mod]),
    )
    return lecture_list_db, trans.get_result(auth_stat), chapters_db, media_db


@api_route("/lecture/<int:lecture_id>", ["GET"], allow_while_readonly=True)
def api_route_lecture(lecture_id: int):
    is_mod: bool = is_moderator()
    lecture_list_db, auth_db, chapters_db, media_db = db_pool.execute_read_transaction(
        _db_execute_get_lecture,
        is_mod,
        lecture_id
    )
    
    if len(lecture_list_db) == 0:
        raise ApiClientException(ERROR_UNKNOWN_OBJECT)
    if len(lecture_list_db) > 1:
        raise Exception("Got more than one lecture with id " + str(lecture_id))  # pragma: no cover
    lecture_db = lecture_list_db[0]
    if not is_mod and (not bool(lecture_db["visible"])
                       or not bool(lecture_db[COURSE_SECONDARY_COLUMN_VISIBLE])):
        raise ApiClientException(ERROR_UNAUTHORIZED)

    course_json = course_secondary_db_to_json_no_lectures(lecture_db, auth_db, is_mod)
    lecture_json = lecture_db_to_json_no_chapters_media(
        lecture_db, auth_db, not bool(lecture_db[COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE]), is_mod)
    lecture_json["chapters"] = list(map(lambda c: chapter_db_to_json(c, is_mod), chapters_db))
    lecture_json["media_sources"] = list(map(
        lambda m: media_source_with_format_db_to_json(m,
                                                      bool(lecture_db[COURSE_SECONDARY_COLUMN_DOWNLOADABLE]), is_mod),
        media_db))
    
    return {
        "course": course_json,
        "lecture": lecture_json
    }


_SQL_EXISTS_LECTURE = PreparedStatement(f"""
SELECT COUNT(*) AS "count"
FROM "lectures"
JOIN "courses" ON ("courses"."id" = "lectures"."course_id")
WHERE "lectures"."id" = ?
    AND ( ? OR (
            "lectures"."visible"
        AND "courses"."visible"
    ))
""")
_SQL_GET_CHAPTER_COUNT = PreparedStatement(f"""
SELECT COUNT(*) AS "count" FROM "chapters"
WHERE "lecture_id" = ?
""")
_SQL_PUT_CHAPTER_SUGGESTION = PreparedStatement(f"""
INSERT INTO "chapters" ("lecture_id", "time", "text", "visible")
VALUES (?, ?, ?, ?)
""")

_CHAPTER_LIMIT_FOR_SUGGESTIONS = api.config["API_CHAPTER_SUGGESTIONS_LIMIT_PER_LECTURE"]
_CHAPTER_SUGGESTIONS_RATE_LIMITER = IntervalRateLimiter(
    api.config["API_CHAPTER_SUGGESTIONS_RATE_LIMIT"]["window_size_seconds"] * pow(10, 9),
    api.config["API_CHAPTER_SUGGESTIONS_RATE_LIMIT"]["max_request_count"],
    api.config["API_CHAPTER_SUGGESTIONS_RATE_LIMIT"]["interval_size_seconds"] * pow(10, 9),
    "chapter_suggestions.rate_limiter"
)


def _db_execute_chapter_suggestion(trans: WriteTransaction, is_mod: bool, lecture_id: int, start_time: int, name: str):
    lecture_count_db, chapter_count_db = trans.execute_statements(
        (_SQL_EXISTS_LECTURE, [lecture_id, is_mod]),
        (_SQL_GET_CHAPTER_COUNT, [lecture_id])
    )
    if lecture_count_db[0]["count"] == 0:
        raise ApiClientException(ERROR_UNKNOWN_OBJECT)
    if chapter_count_db[0]["count"] >= _CHAPTER_LIMIT_FOR_SUGGESTIONS:
        raise ApiClientException(ERROR_TOO_MANY_SUGGESTIONS)
    trans.execute_statement_and_commit(_SQL_PUT_CHAPTER_SUGGESTION, lecture_id, start_time, name, False)


@api_route("/lecture/<int:lecture_id>/chapter_suggestion", ["PUT"])
def api_route_lecture_chapter_suggestion(lecture_id: int):
    is_mod = is_moderator()
    json_request = get_client_json(request)
    start_time = json_request.get_int("start_time", 0, MAX_VALUE_SINT32)
    name = json_request.get_string("name", 128)

    if not _CHAPTER_SUGGESTIONS_RATE_LIMITER.check_new_request():
        raise ApiClientException(ERROR_TOO_MANY_SUGGESTIONS)
    db_pool.execute_write_transaction(_db_execute_chapter_suggestion, is_mod, lecture_id, start_time, name)
    return {}, HTTP_201_CREATED


def _db_execute_search(trans: ReadTransaction, is_mod: bool, search_term: str):
    course_list_statement = course_queue_search(trans, search_term, is_mod)
    lecture_list_statement = lecture_queue_search(trans, search_term, is_mod)
    
    course_list_db = trans.get_result(course_list_statement)
    lecture_list_db = trans.get_result(lecture_list_statement)
    
    if len(course_list_db) == 0 and len(lecture_list_db) == 0:
        trans.close()
        return course_list_db, lecture_list_db, None, None, None, None
    
    course_ids: dict[int, None] = {}
    for course_db in course_list_db:
        course_ids[course_db["id"]] = None
    
    lecture_ids: dict[int, None] = {}
    for lecture_db in lecture_list_db:
        lecture_ids[lecture_db["id"]] = None
        course_ids[lecture_db["course_id"]] = None
    
    assert len(course_ids) > 0
    course_auth_statement = trans.queue_statement(f"""
                    SELECT * FROM "perm"
                    WHERE "course_id" IN ({",".join(map(lambda _: "?", course_ids))})
                    ORDER BY "course_id" ASC
                """, *course_ids)
    if len(lecture_ids) == 0:
        trans.close()
        return course_list_db, lecture_list_db, trans.get_result(course_auth_statement), None, None, None
    
    lecture_where_clause = f""""lecture_id" IN ({",".join(map(lambda _: "?", lecture_ids))})"""
    lecture_auth_statement = trans.queue_statement(f"""
                    SELECT * FROM "perm"
                    WHERE {lecture_where_clause}
                    ORDER BY "lecture_id" ASC
                """, *lecture_ids)
    lecture_chapters_statement = trans.queue_statement(
        _SQL_GET_LECTURE_CHAPTERS_NO_ID_COND % lecture_where_clause, *lecture_ids, is_mod)
    lecture_media_sources_statement = trans.queue_statement(
        _SQL_GET_LECTURE_MEDIA_NO_ID_COND % lecture_where_clause, *lecture_ids, is_mod)
    trans.close()
    return (
        course_list_db,
        lecture_list_db,
        trans.get_result(course_auth_statement),
        trans.get_result(lecture_auth_statement),
        trans.get_result(lecture_chapters_statement),
        trans.get_result(lecture_media_sources_statement),
    )


@api_route("/search", ["GET"], allow_while_readonly=True)
def api_route_search():
    is_mod = is_moderator()
    
    if "q" not in request.args:
        raise ApiClientException(ERROR_REQUEST_MISSING_PARAMETER("URL.q"))
    search_term = request.args["q"]
    if len(search_term) > 1024:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("URL.q", "Too long"))
    if search_term.isspace():
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("URL.q", "Only whitespace is not allowed"))
    
    course_list_db, lecture_list_db, course_auth_db, lecture_auth_db, lecture_chapters_db, lecture_media_sources_db = \
        db_pool.execute_read_transaction(
            _db_execute_search,
            is_mod,
            search_term
        )

    if len(course_list_db) == 0 and len(lecture_list_db) == 0:
        return {
            "courses": [],
            "lectures": [],
            "courses_context": {}
        }
    
    course_auth_by_id = db_group_data_by_id(course_auth_db, "course_id")
    
    courses_context_json: dict[str, JsonTypes] = {}
    course_id_list_json: list[int] = []
    for course_db in course_list_db:
        id = course_db["id"]
        course_id_list_json.append(id)
        course_auth = course_auth_by_id.get(id, [])
        courses_context_json[str(id)] = course_db_to_json_no_lectures(course_db, course_auth, is_mod)
    
    if len(lecture_list_db) == 0:
        return {
            "courses": course_id_list_json,
            "lectures": [],
            "courses_context": courses_context_json
        }
    
    lecture_auth_by_id = db_group_data_by_id(lecture_auth_db, "lecture_id")
    chapters_by_id = db_group_data_by_id(lecture_chapters_db, "lecture_id")
    media_sources_by_id = db_group_data_by_id(lecture_media_sources_db, "lecture_id")
    
    lecture_list_json = []
    for lecture_db in lecture_list_db:
        id = lecture_db["id"]
        course_id = lecture_db["course_id"]
        course_auth = course_auth_by_id.get(course_id, [])
        lecture_auth = lecture_auth_by_id.get(id, [])
        
        lecture_json = lecture_db_to_json_no_chapters_media(
            lecture_db,
            course_auth + lecture_auth,
            not bool(lecture_db[COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE]),
            is_mod
        )
        lecture_json["chapters"] = list(map(lambda c: chapter_db_to_json(c, is_mod), chapters_by_id.get(id, [])))
        lecture_json["media_sources"] = list(map(
            lambda m: media_source_with_format_db_to_json(m, lecture_db[COURSE_SECONDARY_COLUMN_DOWNLOADABLE], is_mod),
            media_sources_by_id.get(id, [])))
        lecture_list_json.append(lecture_json)
        
        if str(course_id) not in courses_context_json:
            courses_context_json[str(course_id)] = course_secondary_db_to_json_no_lectures(lecture_db, course_auth, is_mod)
    
    return {
        "courses": course_id_list_json,
        "lectures": lecture_list_json,
        "courses_context": courses_context_json
    }
